/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/******************************************************************************
 program name: lm_ibdtests.c  ( Author: Saonli Basu)

 Program to perform trait-model-free analysis for linkage detection. The
 program performs three diffferent tests for linkage:
    1) normality-based test
    2) permutation test
    3) likelihood-ratio test.
 On extended pedigrees, the IBD configurations of affected and/or unaffected
 individuals conditional on the marker data are sampled using MCMC. Under
 normality-based test and/or permutation test, four different IBD measures
 (S_{pairs}, S_{robdom}, S_{m\lambda}, S_{au} can be used to score a sampled
 IBD configuration in the program. Under likelihood-based approaches, the
 likelihood-ratio statistic under `delta' model and/or `(lambda,p)' model can
 be scored for each sampled IBD configuration. The final statistic is taken as
 a mean of the scores over N MCMC scans. The test for linkage is performed
 using the permutation Test, normality-based test or likelihood-ratio test.

*******************************************************************************/
/**
 * @defgroup lm_auto
 * @ingroup autozyg
 * @{
 */

#define PNAM  "lm_ibdtests"
#include "autozyg.h"
#include "auto_chk.c"
#include "./Utils/ibdtests.h"
#include "./Utils/mem_utility.c"
#include "./Utils/permutation.c"
#include "./Utils/permutation_sibship.c"
#include "./Utils/cdfcompute.c"
#include "./Utils/prntIBDT.c"
#include "./Utils/apputil.c"
#include "./Utils/lambdamodel.c"
#include "./Utils/read_write.c"
#include "./Utils/statistics.c"
#define nMaxFileFD 200
#define AUTO_CKPNTS 1000
#define SEG_LAMBDA 9

/*****  Global Variables for Allele Sharing Variables ******************/
int ***matry = NULL;  /** stores nghd[i].index from all permutations **/
/**** Allocating arrays for choosing sibships for permutation ***/
int **size_aff = NULL;
int **size_unaff = NULL;
int  *Group = NULL;
int **Size = NULL;
int **Affect_resimulation = NULL;

int main (int argc, char *argv[])
{
   int   i, j, t, c;
   int   nNoOfStatOfInts;
   int   nsw;
   int **count;                         /* [nc][nl]                       */
   double ****nscore;                   /* [nc][ns][nl][pt]  ibd counts:
                                         *   ea component, ea scoreset,
                                         *   ea locus, ea pattern         */
   int   tot_count = 0;                 /* score by component and locus   */

   int ***recounts;                     /* [nc][nl][6] counts of recombs  */
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **prob;
   int   chkitvl = 0;                   /* For AUTO_CKPNTS when this is set.  */
   FILE *oSeedF;
   Measure *score_aff;
   int   locus;
   int   iter1;
   int   trt_data_Tag = 0;
   int   Total_Sampling = 0;
   FILE *write_fp = NULL;
   char  pStrPermFile[MAX_FILE_PATH];
   char  pStrObsFile[MAX_FILE_PATH];
   int   retVal = 0;
   /*** Variables for p value computation for Normality Test ***/
   int   bExecutionContinue = 0;
   line_info_struct *histogramDataPtr = NULL;

   /******************************/
   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();

   /* set up permutation arrays for loci and switches  if Scan (not Step) */
   /* proc_seeds() has already assigned the permutation seeds */
   /* (2 ways, either from par file or default value) */
   /*********************************/

   trt_data_Tag = 0;

   for (i = 0; i < pedsize; i++)
   {
      if (nghd[i].d_dat != NULL)
      {
         trt_data_Tag = 1;
         break;;
      }
   }
   /**** Selecting individuals for the permutation test
         and writing them in a temporary file *****/
   if (PermutEnable)
   {
      int   return_val = 0;

      NPermutIt += 1;

      sprintf (pStrPermFile, "./permutation.txt");
      sprintf (pStrObsFile, "./observed.txt");

      return_val = Write_Permutation (pStrPermFile, pStrObsFile);
      if (return_val == 0)
      {
         goto ERROR_LEVEL_2;
      }
   }
   initialize_variables (SEG_LAMBDA, NMarks); /* Initializing variables */

   /** Generating permutations for the permutation test **/
   if (PermutEnable)
   {
      if (!rand_permutation (NPermutIt, Total_Size, pStrPermFile, pStrObsFile))
      {
         printf ("Error exitting from rand_permutation at line (%d)\n",
                 __LINE__);
         goto ERROR_LEVEL_2;
      }
   }
   /** computing weights for Slambda measure ****/
   if (LambdaOn)
   {
      Weight_Multilambda (paff);
   }
   /***assigning different IBD measures to score_aff variable ***/
   if (NormalEnable || PermutEnable)
      initialize_score (&score_aff);
   if (LambdaOn)
      initialize_valarray (paff, SEG_LAMBDA);
   /*************************************************/
   if (PermutEnable)
   {
      /* Saves the ending permutation seeds */
      getSeeds (PERM);

      /* Assigns the sampling seed values,
         (2 ways, either from par file or ending permutation seeds) */
      assgnSeeds (SAMPL);
      setSeeds (SAMPL);
   }

   /* includes initializing seeds for rans() */
   prob = setupInherit (1);
   tot_count = 0;

   /* Try to restart if a rescue file is specified.                           */
   t = 0;
   auto_restart (nscore, tot_log_prob_pen, tot_log_prob_seg, prob,
                 count, recounts, &tot_count, &t);

   prntStatusBMC ();

   /*** computing null mean and null variance for different
        allele-sharing statistics ***/
   if (NormalEnable || ChiSqEnable)
   {
      Calculate_Stat (nsw);
   }
   /* now do burn-in iterations */
   printf (" Now starting %d iterations of burn-in\n", NBrnIt);
   for (t = 1; t <= NBrnIt; t++)
   {
      if (!(t % ChkIt))
      {
         printf (" Starting burn-in iteration %d\n", t);
         fflush (stdout);
      }
      sampling (t, 1, NMarks, nsw, nFGL, prob);
   }

   /* now do MCMC sampling, and score results */

   for (t = 0; t < NIt; ++t)
   {
      /* MCMC sampling step or scan */
      sampling (t, 1, NMarks, nsw, nFGL, prob);

      /*******************  Scoring different IBD measures *****************/
      if (((t + 1) % ScorIt) == 0)
      {
         /*** Computation of IBD Measures for all permutations
              and for all loci ***/
         if (PermutEnable)
         {
            reset_measurevars (NMarks);
         }
         for (c = 0; c < n_compnts; c++)
         {
            if (NormalEnable)
            /*** computing IBD measures for each MCMC scan
                 for normality-based test ****/
            {
               for (locus = 1; locus < (NMarks + 1); locus++)
               {
                  assign_values (c, locus);
               }
            }
            if (ChiSqEnable)
            /**** computing statistics for each MCMC scan
                   for likelihood-based approaches ***/
            {
               if (Lambda_pOn)
               {
                  Likelambdap_Computation (c, NMarks, IntervalSegL,
                                           IntervalSegP);
               }
               if (DeltaOn)
               {
                  Likedelta_Computation (c, NMarks);
               }
            }
            if (PermutEnable)
            /*** computing IBD measures for each MCMC scan
                 for permutation test ***/
            {
               iter1 = 0;
               while (iter1 < NPermutIt)
               {
                  if (PermutEnable)
                  {
                     int   countv = 0;
                     exint[c][0] = tot_affected[c];
                     exint[c][1] = Total_Size[c] - tot_affected[c];
                     while (countv < tot_affected[c])
                     {
                        exint[c][countv + 2] = matry[c][iter1][countv];
                        if (SAffUnaffOn)
                           exint1[c][(2 * Group[c] + 1) + countv] =
                              matry[c][iter1][countv];
                        countv = countv + 1;
                     }
                     while (countv < Total_Size[c])
                     {
                        exint[c][countv + 2] = matry[c][iter1][countv];
                        if (SAffUnaffOn)
                           exint1[c][(2 * Group[c] + 1) + countv] =
                              matry[c][iter1][countv];
                        countv = countv + 1;
                     }
                  }
                  for (locus = 1; locus < (NMarks + 1); locus++)
                  {
                     reassign_values (c, iter1, locus);
                  }
                  iter1 = iter1 + 1;
               }
            }
            if ((c == (n_compnts - 1)) && (PermutEnable))
            {
               file_write (Total_Sampling, NMarks);
            }
         }
         Total_Sampling = Total_Sampling + 1;
      }
      /*************Done with IBD Measures*************************************/

/* The following code was copied from lm_auto.c so that a call to auto_dump   */
/* would be included here in lm_ibdtests.c.  So long as chkitvl is equal to 0 */
/* the call to auto_dump will not actually occur.  This code was inserted to  */
/* make lm_ibdtests.c be in compliance with the c89 standard for C programs.  */
/* Without this code the C compiler produces a warning that auto_dump is      */
/* defined but never used.  We may possibly at some point in the future wish  */
/* to add the rescue file capability to lm_ibdtests.  Then we would want the  */
/* call to auto_dump to actually occur every chkitvl iterations.              */

      if (chkitvl && ((t + 1) % chkitvl == 0))
      {
         auto_dump (nscore, tot_log_prob_pen, tot_log_prob_seg, prob,
                    count, recounts, tot_count, t);
      }
   }

   bExecutionContinue = 1;
   /* Freeing memory in ERROR_LEVEL */
   goto ERROR_LEVEL_2;

   prntStatusFMC ();

 CALCULATE_PVALUE:
   nNoOfStatOfInts = SpairsOn + SAffUnaffOn + RobdomOn + LambdaOn;
   if ((PermutEnable))
   {
      retVal =
         PerformHistogramAnalysisNew (Total_Sampling, nMaxFileFD, NPermutIt,
                                      NMarks, nNoOfStatOfInts);
      if (retVal == 0)
      {
         printf ("ERROR: Some Error occured in Histogram Analysis,"
                 " Exiting unsuccesfully..\n");
         return 0;
      }
   }
   if (ChiSqEnable)
   {
      /* maximizing likelihood under (lambda,p) model */
      if (Lambda_pOn)
      {
         Max_lambdap (NMarks, Total_Sampling, IntervalSegL, IntervalSegP);
      }
      /* maximizing likelihood under `delta' model */
      if (DeltaOn)
      {
         Max_Delta (NMarks, Total_Sampling);
      }
   }
   if (PermutEnable)
      ReadInputFileData (&histogramDataPtr, Total_Sampling);

   if (PermutEnable)
   {
      prntIBDTHead (PERMUTATION);
      prntIBDTColPerm (SpairsOn, RobdomOn, SAffUnaffOn, LambdaOn, MaxFGLOn);
      for (i = 1; i <= (NMarks); i++)
      {
         for (j = 0; j < nNoOfStatOfInts; j++)
         {
            /* p-value computation under permutation test */
            calculate_p_value_new (histogramDataPtr, i, j + 1, NMarks,
                                   NPermutIt, nNoOfStatOfInts, Total_Sampling,
                                   bResimuEnable, write_fp, nStatType,
                                   pQuantileList);
         }
         printf ("\n");
      }
   }
   if (histogramDataPtr != NULL)
   {
      freeLineList (&histogramDataPtr);
   }
   /* p-value computation for normality-based test */
   if (NormalEnable)
   {
      prntIBDTHead (NORMAL);
      prntIBDTColNorm (SpairsOn, RobdomOn, SAffUnaffOn, LambdaOn, MaxFGLOn);

      Norm_Compute (NMarks, Total_Sampling);
   }
   fflush (stdout);
   /* p-value computation for likelihood-based approaches */
   if (ChiSqEnable)
   {
      if (Lambda_pOn)
         printf ("\n The correlation coefficient for the pedigree data,\n"
                 "  used in the lambda_p model is %5.3f\n", CORR);
      prntIBDTHead (CHISQUARE);
      prntIBDTColChiSq (DeltaOn, Lambda_pOn);

      pVal_Computation (NMarks);
   }
   /****removing temporary files ****/
   remove ("Histogram.txt");
   remove ("./permutation.txt");
   remove ("./observed.txt");
   /*  THIS changed to hard file-names for now -- EAT 8/2008
    *  to avoid valgrind error.
    remove (pStrPermFile);
    remove (pStrObsFile);
    */
   exit (EXIT_SUCCESS);

   /*******free memory ***/
 ERROR_LEVEL_2:
   {
      free_memory (NMarks);
   }
   free_rescue_filenames (orescName);
   free_xxgg ();
   free_allocate_loci ();
   free_gen_pen ();
   free_seg_pen (NMarks + 1, &prob);
   free_logprob ();
   free_msmpl ();
   free_traits ();
   free_ibd_measures ();
   getSeeds (SAMPL);
   if (SamplSeedsOut || PermSeedsOut)
   {
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      /* Writes sampler and/or permutation seed */
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
   if (bExecutionContinue == 1)
   {
      goto CALCULATE_PVALUE;
   }
   else
   {
      exit (EXIT_FAILURE);
   }
}
