/***************************************************************************/
/*  Program to do a basic simulation and analysis of IBD inference with LD */
/*  Routines to process input data and make haplotypes                     */
/*  Contains:  input_haps ():  reads in the initial haplotypes             */
/*             make_haplo ():  recycles and subselects SNPs for analysis   */
/*           make_haplo_segs():  copies segments from constr ibd to haplos */
/*           make_data():  set the haplo data from constructed haplos      */
/*           select_snps(): iselects SNPs for analysis from input block    */
/*           print_haplo(): prints the snp data on a haplotype             */
/***************************************************************************/

void input_haps (FILE * fpt, int mhap, int mst, int mgap, int mblk,
                 int msnp, int **hap, int *dex, int totmark)
/* reads in haplotypes from input file;
      sets a block of mblk SNPS at gaps of mgaps SNPs    */
{
   int   i, k, j, jjj, jj, dummy;
   for (i = 0; i < mhap; ++i)
   {
      for (j = 0; j < mst; ++j)
         fscanf (fpt, "%d", &dummy);
      k = 0;
      for (jj = 0; jj < msnp; ++jj)
      {
         for (jjj = 0; jjj < mblk; ++jjj)
         {
            fscanf (fpt, "%d", &(hap[i][k]));
            dex[k] = j + 1;
            ++j;
            ++k;
         }
         for (jjj = 0; jjj < mgap; ++jjj)
         {
            fscanf (fpt, "%d", &dummy);
            ++j;
         }
      }
      for (jjj = j; jjj < totmark; ++jjj)
         fscanf (fpt, "%d", &dummy);
   }
   return;
}

void make_haplo (int *hap, int **hapall, int nmrk, int nst1, int nc, int nsnp)
/*  make a 10000 (nmrk) length haplo by combining up to nc of the nsnp-length
       haplos starting at haplo  nst1   (nc=5, nsnp=2000)    */
{
   int   i, j, k, mk;
   j = 0;
   k = nst1;
   for (i = 0; i < nc && j < nmrk; ++i)
   {
      for (mk = 0; mk < nsnp; ++mk)
      {
         hap[j] = hapall[k][mk];
         ++j;
      }
      ++k;
   }

   return;

}

void make_haplo_segs (int *hap, int nmrk, Chromseglist * chromo,
                      int **hapset, double mksp)
/* copy the relevant segments of founder chromosomes to the data haplotype,
    at marker spacing mksp (in cM) for total nmrk markers */
/* modified so that data are now 1,2 instead of 0,1       */
{
   int   i;
   double dd;
   Chromseglist *cc;
   i = 0;                               /* current marker index */
   dd = 0.0;                            /* current position     */

   cc = chromo;
   while (cc)
   {
      while (dd < cc->current->endpos)
      {
         hap[i] = hapset[(cc->current->fgl) - 1][i] + 1;
         ++i;
         dd = mksp * i;
      }
      cc = cc->next;
   }
   return;
}

void make_pair_data (int *hd, int **hm, int nd, int skp, int stmk, int *prpr)
/* set data for pair of haplotypes (prpr) only  */
{
   int   i, j, k1, k2;
   k1 = prpr[0];
   k2 = prpr[1];
   for (i = 0; i < nd; ++i)
   {
      j = i * skp + stmk;               /* setting up marker j */
      hd[i] = -1;
      if ((hm[k1][j] == 1) && (hm[k2][j] == 1))
         hd[i] = 0;
      if ((hm[k1][j] == 1) && (hm[k2][j] == 2))
         hd[i] = 1;
      if ((hm[k1][j] == 2) && (hm[k2][j] == 1))
         hd[i] = 2;
      if ((hm[k1][j] == 2) && (hm[k2][j] == 2))
         hd[i] = 3;
   }
   return;
}

void make_data (int *hd, int **hm, int nh, int kk, int nd, int skp, int stmk)
/* make data length nd, from kk'th quartet of haplos
   using every skp'th marker from nh haplotypes hm  */
/*  allow for missing data --- hd =-1 if any hm=0  */
/*  igdx is (now Global) genotypic data indicator for analysis */
{
   int   i, j, k;
   for (i = 0; i < nd; ++i)
   {
      j = i * skp + stmk;               /* setting up marker j */
      hd[i] = 0;
      for (k = kk * nh; k < nh + kk * nh; ++k)
      {
         if (hm[k][j] == 0)
         {
            hd[i] = -1;
            break;
         }
         hd[i] = 2 * hd[i] + (hm[k][j] - 1);
      }
   }
   if (igdx)                            /* reduce from 16 = 2^4 haps, to 9 genos */
   {
      for (i = 0; i < nd; ++i)
      {
         if ((hd[i] == 1) || (hd[i] == 2))
            hd[i] = 1;
         if ((hd[i] == 3) || (hd[i] == 4))
            hd[i] = hd[i] - 1;
         if ((hd[i] == 5) || (hd[i] == 6) || (hd[i] == 9) || (hd[i] == 10))
            hd[i] = 4;
         if ((hd[i] == 7) || (hd[i] == 11))
            hd[i] = 5;
         if (hd[i] == 8)
            hd[i] = 3;
         if (hd[i] == 12)
            hd[i] = 6;
         if ((hd[i] == 14) || (hd[i] == 13))
            hd[i] = 7;
         if (hd[i] == 15)
            hd[i] = 8;
      }
   }
   return;
}

int select_snps (int **hap1, int **hap2, int mhap, int mblk, int msnp,
                 double *pif, int *dex1, int *dex0, double qq)
{
   int   i, j, k, kk, cc;

   kk = 0;
   for (k = 0; k < msnp; ++k)
   {
      for (j = k * mblk; j < (k + 1) * mblk; ++j)
      {
         cc = 0;
         for (i = 0; i < mhap; ++i)
            cc += hap1[i][j];
         if ((cc < (1 - qq) * mhap) && (cc > qq * mhap))
            break;
      }
      if (j < (k + 1) * mblk)
      {
         for (i = 0; i < mhap; ++i)
            hap2[i][kk] = hap1[i][j];
         dex0[kk] = dex1[j];
         pif[kk] = cc / (mhap + 0.1 - 0.1);
         /*              printf ("\n %d %f ",kk, pif[kk]);
            for (i=0; i<mhap; ++i) printf("%d",hap2[i][kk]);  */
         ++kk;
      }
      /*        else
         printf ("\n no snps in block %d",k);     */
   }
   printf ("\n number of selected snps is %d", kk);
   return (kk);
}

void print_haplo (FILE * fptr, int *hap, int mrk)
{
   int   j;
   fprintf (fptr, "\n");
   for (j = 0; j < mrk; ++j)
      fprintf (fptr, "%2d", hap[j]);
   return;
}
