/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 AUTHOR
    Written by Yun Ju Sung.

 SEE ALSO
    README_twoqtl (in TwoQTL directory) for user documentation.
    Gold_twoqtl directory to see examples.

 HISTORY
    Initial codes started from lm_markers.c (at 11/11/04).
    About 4000 lines of code (mostly in ../TwoQTL).  Apr  5, 2005
    Extended to do componentwise.                    May 20, 2005
    Cvs commit done.                                 Jul  5, 2005

    Use two streams of random numbers,
    one for segregation indicator and
    the other for founder gene alleles.              Aug  8, 2005

    Evaluate the null position more
    (to get more precision).                         Aug 18, 2005

 BUGS
    (1) sampling founder gene needs testing (more?)
    (2) need to get parameter statements
    (3) proper error messages
*/

/**
 * @defgroup lm_twoqtl lm_twoqtl
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_twoqtl"
#define TWOQTL

#include "lodscore.h"
#include "twoqtl.h"

double *set_theta ();
void  finishLodSc (void);

int main (int argc, char *argv[])
{
   mc_eval_t *MC_eval, *mc_eval;        /* MC evaluations for each batch.     */

   char ***S_trait_p;                   /* Paternal SI at trait positions.    */
   char ***S_trait_m;                   /* Maternal SI at trait positions.    */

   double **prob = NULL;
   double *theta = NULL;

   int   tot_count;                     /* Total number of iterations to be   */
   /*                                      used for lod score estimation.     */
   /*                                      For example, if 1000 iterations    */
   /*                                      are scored every 20, then          */
   /*                                      tot_count = 50.                    */

   int   bc;                            /* Current batch indicator.           */

   int   nsw, i_max, mc_size, start;
   int   iteration, mc_eval_offset;
   int   i1, i2;

   /* ................................  SET-UP   ............................ */

   nsw = setupAutozyg (argc, argv, PNAM, 2);
   setupSampling ();

/* Pass NTLocs as the argument to setupInherit.  This will be passed on as    */
/* the mm variable throughout the rest of the setup procedure.                */

   prob = setupInherit (NTLocs);

/* Copy the validated additive and residual variances to the SigAdd and       */
/* SigErr global variables, respectively, so that the caller's values for     */
/* these components of variance will be available throughout lm_twoqtl.       */

   SigAdd = P_trait_pars->addi_var;
   SigErr = P_trait_pars->resi_var;

/* We also need to initialize the lm_twoqtl variables holding the mean trait  */
/* values for each tloc for each genotype: geno_means_1[] and geno_means_2[]. */

   geno_means_1 = P_trait_pars->p_tloc_pars[0].means;
   if (NQtl == 2)
      geno_means_2 = P_trait_pars->p_tloc_pars[1].means;

   /* ...    Parameters (more later).    ...                                  */
   set_twoqtl_parm ();                  /* Note: This call is temporary.      */

   theta = set_theta ();

   tot_count = NIt / ScorIt;
   if (tot_count * ScorIt != NIt)
      give_warn (ERR_err,
                 "Scoring interval does not divide realizations count");

   batch_size = tot_count / NBtch;
/* Check that the number of batches divides the number of scored iterations.  */
   if (batch_size * NBtch != tot_count)
      give_warn (ERL_bt1, NBtch);

/* Make sure that batch_size is a positive number.                            */
   if (batch_size <= 0)
      give_up (ERL_bt2, NBtch);

   printf (" %d realizations will be scored, at scoring interval %d\n",
           tot_count, ScorIt);
   printf (" Scored realizations will be batched in %d sets of size %d\n",
           NBtch, batch_size);
   printf ("\n\n");
   fflush (stdout);

   MC_eval = alloc_mc_eval (NLocPos);

   ibase = get_ibase ();                /* Index of first ind in each comp.   */
   size_c = get_size_c ();              /* Number of ind in each component.   */
   obs_index = get_obs_index ();

   S_trait_p = alloc_S_trait (NLocPos, size_c);
   S_trait_m = alloc_S_trait (NLocPos, size_c);

   mbase = get_mbase (nsw);
   n_meiosis = get_n_meiosis (nsw);

   prntStatusBMC ();

   if (debug_level == 6)
   {
      i_max = n_compnts * NBtch;
      for (i1 = 0; i1 < i_max; ++i1)
         print_mc_eval (MC_eval[i1]);
   }

   /* ................  BURN-IN  .......................................      */

   printf (" Now starting %d iterations of burn-in\n", NBrnIt);

   for (iteration = 0; iteration < NBrnIt; ++iteration)
   {
      if (!(iteration % ChkIt))
      {
         printf (" Starting burn-in iteration %d\n", iteration);
         fflush (stdout);
      }
      sampling (iteration, NTLocs, (NLoci - 1), nsw, nFGL, prob);
   }
   printf ("\n");
   fflush (stdout);

   set_evaluate ();                     /* Initialize various global          */
   /*                                      variables used by evaluate.        */

   /* ................  COMPUTE USING MCMC  ............................      */
   mc_size = 0;
   start = 1;
   bc = 0;

   for (iteration = 0; (iteration < NIt && bc < NBtch); ++iteration)
   {
      /* Update xx1, xx0.                                                     */
      sampling (iteration, NTLocs, (NLoci - 1), nsw, nFGL, prob);

      if (debug_level == 2)
         dbgXXGG ();

      if (!((iteration + 1) % ScorIt))
      {
         mc_eval_offset = n_compnts * bc;
         for (i1 = 0; i1 < n_compnts; ++i1)
         {
            update_S_trait (i1, nsw, S_trait_p, S_trait_m);

            /* Update batch sums at the alternative and null hypotheses.      */
            mc_eval = MC_eval + mc_eval_offset + i1;
            log_prob_alt (theta, 0, S_trait_p, S_trait_m, i1, nsw, start,
                          mc_eval);

            /* Update for the null hypothesis.                                */
            for (i2 = 0; i2 < MultiplierForNull; ++i2)
               log_prob_null (theta, 0, i1, nsw, start && (i2 == 0), mc_eval);

            if (debug_level > 3)
               print_mc_eval (MC_eval[mc_eval_offset + i1]);
         }

         ++mc_size;

         /* Write batch means and start new batch if current batch is done.   */
         if (!(mc_size % batch_size))
         {
            write_batch_means ((MC_eval + n_compnts * bc), bc);
            start = 1;
            ++bc;
         }
         else
            start = 0;
      }
   }

   /* Compute lod scores and write them to the standard output file.          */
   write_lod_score (MC_eval);

   /* Free up allocated memory.                                               */
   free_update_S_trait ();
   if (SigAdd > 0.0)
      free_tq_gauss ();
   free_component ();
   free_evaluate ();

   free_S_trait (S_trait_p);
   free_S_trait (S_trait_m);

   free_mc_eval (MC_eval);
   free (theta);
   finishLodSc ();

   exit (EXIT_SUCCESS);
}

double *set_theta ()
{
   double *theta;
   int   n_theta;

   n_theta = NQtl * 4 + 2;
   theta = (double *) malloc (n_theta * sizeof (double));
   alloc_chk (WHERE, theta);

   theta[0] = SigErr;
   theta[1] = SigAdd;
   theta[2] = Qq[0][1];
   memcpy (theta + 3, geno_means_1, 3 * sizeof (double));

   if (NQtl == 2)
   {
      theta[6] = Qq[1][1];
      memcpy (theta + 7, geno_means_2, 3 * sizeof (double));
   }

   return theta;
}

void finishLodSc (void)
{
   FILE *oSeedF;                        /* File pointer to the seed file.     */

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   /* Free up allocated memory.                                               */
   free_xxgg ();
   free_allefrq ();
   free_chromo (0);
   free_allocate_loci ();
   free_perm ();
   locus_peel_free ();

   return;
}
