/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*
 * gauss_quant.c - May 23, 1997 (SCH)
 *
 * Likelihood evaulation for polygenic trait by gaussian elimination.
 */

#include "quant.h"
#include "peelie.h"

/* Sparse matrix element (Numerical Recipes, Press et al.).  Only useful if you
 * know the structure of the matrix (with regard to non-zero elements) before
 * you start.  Used here to store the inverse of the G matrix.                */

struct sp_element
{
   int   x;
   double val;
};

/* For building up contribution from an individual.                           */
struct rvec
{
   int   i;
   double val;
};

static int *sire_list, *dam_list, *data_index, *eq_pos, psize, nfam;
static int *recode = 0, *fam_list = 0;

static struct sp_element *AIMatrix = 0;
static struct Off *BBlock = 0, *freelist = 0; /* Off-Diagonal Elements.       */
static struct Off *BaseBlock = 0;

int   bp = 0;                           /*        Block pointer, block size. */

/* Recode individuals so parents are (numerically) before offspring.  A lot   */
/* of the gauss_quant program assumes this.                                   */

void set_code (int i, int *id)
{
   int   ids, idd;

   ids = nghd[i].pa->index;
   idd = nghd[i].ma->index;

   if (ids != INVALID_INDEX && !recode[ids])
      set_code (ids, id);
   if (idd != INVALID_INDEX && !recode[idd])
      set_code (idd, id);

   if (ids != INVALID_INDEX)
      sire_list[*id] = recode[ids];
   if (idd != INVALID_INDEX)
      dam_list[*id] = recode[idd];
   data_index[*id] = qdex[i];
   recode[i] = ++(*id);
}

/* Calculate inverse of NRM (G) matrix using Quaas/Henderson method.  See     */
/* Numerical Recipes in C 2nd edition pages 78-79 for a description of the    */
/* sparse matrix storage.  Note the matrix is half stored.                    */

void calc_nrm (int id)
{
   int   i, j, k, ids, idd, ids1, idd1, nz, *temp;
   double *u, *v, xx, d, dt = 0.0, d2, d4;
   Marlist *ml;

   temp = malloc (id * sizeof (int));
   alloc_chk (WHERE, temp);

   /* Count non-zero (off diagonal) elements in matrix. */
   for (i = nz = 0; i < pedsize; i++)
      if ((j = recode[i]))
      {
         j--;
         temp[j] = i;
         if ((ids = sire_list[j]))
            nz++;
         if ((idd = dam_list[j]))
            nz++;
         ml = nghd[i].marriages;
         while (ml)
         {
            k = recode[ml->current->spouse->index] - 1;
            if (j > k)
               nz++;
            ml = ml->next;
         }
      }
   AIMatrix = malloc ((nz + 1 + id) * sizeof (struct sp_element));
   alloc_chk (WHERE, AIMatrix);

   /* Set up coordinates for non-zero elements. */
   nz = id + 1;
   for (j = 0; j < pedsize; j++)
   {
      i = temp[j];
      AIMatrix[j].val = 0.0;
      AIMatrix[j].x = nz;
      ml = nghd[i].marriages;
      if ((ids = sire_list[j]))
      {
         AIMatrix[nz].x = ids - 1;
         AIMatrix[nz++].val = 0.0;
      }
      if ((idd = dam_list[j]))
      {
         AIMatrix[nz].x = idd - 1;
         AIMatrix[nz++].val = 0.0;
      }
      ml = nghd[i].marriages;
      while (ml)
      {
         k = recode[ml->current->spouse->index] - 1;
         if (j > k)
         {
            AIMatrix[nz].x = k;
            AIMatrix[nz++].val = 0.0;
         }
         ml = ml->next;
      }
   }
   AIMatrix[id].x = nz;
   free (temp);
   temp = NULL;

   u = malloc (id * 2 * sizeof (double));
   alloc_chk (WHERE, u);

   v = u + id;
   for (i = 0; i < id; i++)
      u[i] = 0.0;

   /* Build up matrix. */
   for (i = 0; i < id; i++)
   {
      xx = 0.0;
      ids = sire_list[i] - 1;
      idd = dam_list[i] - 1;
      if (ids >= 0)
         xx = u[ids];
      if (idd >= 0)
         xx += u[idd];
      xx = 1.0 - .25 * xx;
      d = 1.0 / xx;
      u[i] += xx;
      xx = sqrt (xx);
      v[i] = xx;
      dt += log (xx);
      for (k = i + 1; k < id; k++)
      {
         xx = 0.0;
         ids1 = sire_list[k] - 1;
         idd1 = dam_list[k] - 1;
         if (ids1 >= i)
            xx = v[ids1];
         if (idd1 >= i)
            xx += v[idd1];
         if (xx > 0.0)
         {
            xx *= .5;
            u[k] += xx * xx;
         }
         v[k] = xx;
      }
      AIMatrix[i].val += d;
      d2 = -.5 * d;
      d4 = .25 * d;
      j = AIMatrix[i].x;
      if (ids >= 0)
      {
         AIMatrix[j++].val += d2;
         AIMatrix[ids].val += d4;
      }
      if (idd >= 0)
      {
         AIMatrix[j++].val += d2;
         AIMatrix[idd].val += d4;
         if (ids >= 0)
         {
            if (ids == idd)
               AIMatrix[ids].val += d4;
            else if (ids > idd)
            {
               for (j = AIMatrix[ids].x; j < AIMatrix[ids + 1].x; j++)
                  if (AIMatrix[j].x == idd)
                  {
                     AIMatrix[j].val += d4;
                     break;
                  }
            }
            else
            {
               for (j = AIMatrix[idd].x; j < AIMatrix[idd + 1].x; j++)
                  if (AIMatrix[j].x == ids)
                  {
                     AIMatrix[j].val += d4;
                     break;
                  }
            }
         }
      }
   }
   dt += dt;
   /* Determinant of matrix */
   AIMatrix[id].val = dt;
   free (u);
   u = NULL;
}

/* Setup data structures for gauss_quant.  This subroutine is only called     */
/* once, by gauss_quant.                                                      */

void gauss_setup (int dom)
{
   int   i, j, k, id = 0;
   Marlist *ml;
   Indlist *kid;

   recode = calloc (4 * pedsize, sizeof (int));
   alloc_chk (WHERE, recode);
   sire_list = recode + pedsize;
   dam_list = sire_list + pedsize;
   data_index = dam_list + pedsize;
   alloc_individual_elements = N;

   /* Make sure that the BMatrix pointer is initialized to NULL.              */
   BMatrix = NULL;

   /* Recode pedigree so parents come before offspring. */
   for (i = 0; i < pedsize; i++)
      if ((!recode[i]) && (nghd[i].index != INVALID_INDEX))
         set_code (i, &id);

   /* Calculate inverse of NRM (G) matrix. */
   calc_nrm (id);

   nfam = 0;
   if (dom)
   {
      /* If dominance effect required, count how many levels of effect we need.  */
      fam_list = calloc (id, sizeof (int));
      alloc_chk (WHERE, fam_list);
      for (i = 0; i < pedsize; i++)
         if ((j = recode[i]))
         {
            ml = nghd[i].marriages;
            while (ml)
            {
               if (j > recode[ml->current->spouse->index])
               {
                  nfam++;
                  kid = ml->current->kids;
                  while (kid)
                  {
                     fam_list[recode[kid->current->index] - 1] = nfam;
                     kid = kid->next;
                  }
               }
               ml = ml->next;
            }
         }
      for (i = 0; i < id; i++)
         if (!fam_list[i] && data_index[i])
            fam_list[i] = ++nfam;

      /* Set up order of equations so the dominance effect for a family       */
      /* is eliminated after the last offspring from that family.             */

      eq_pos = calloc (id + nfam, sizeof (int));
      alloc_chk (WHERE, eq_pos);
      k = 0;
      for (i = 0; i < id; i++)
      {
         if ((j = fam_list[i]))
         {
            j += id - 1;
            if (!eq_pos[j])
               eq_pos[j] = ++k;
         }
         eq_pos[i] = ++k;
      }
   }
   else
   {
      eq_pos = malloc (id * sizeof (int));
      alloc_chk (WHERE, eq_pos);
      for (i = 0; i < id; i++)
         eq_pos[i] = i + 1;
   }
   psize = id;
}

/* Off diagonal elements for the likelihood calculation are allocated in      */
/* blocks.                                                                    */

void AllocBlock (void)
{
   if (!BBlock->Next)
   {
      BBlock->Next = malloc (psize * sizeof (struct Off));
      alloc_chk (WHERE, BBlock->Next);
      BBlock->Next->Next = NULL;
   }
   BBlock = BBlock->Next;
   bp = 1;
}

/* Find storage for another matrix element.  If this subroutine has been      */
/* called from tq_gauss_quant each matrix element is expected to be allocated */
/* individually, without using the freelist.  If this subroutine has been     */
/* called from the original gauss_quant the matrix elements are not allocated */
/* individually and are allocated from the freelist if possible.              */

struct Off *AllocElement (void)
{
   struct Off *p;

   if (alloc_individual_elements)
   {
      p = (struct Off *) malloc (sizeof (struct Off));
      alloc_chk (WHERE, p);

      return p;
   }

   if ((p = freelist))
   {
      freelist = p->Next;
      return p;
   }

   if (bp == psize)
      AllocBlock ();

   return &BBlock[bp++];
}

/* Add x to off-diagonal element r,c.  If the element does not exist yet,     */
/* call AllocElement() to allocate memory for it.                             */

void AddElement (int r, int c, double x)
{
   int   temp;
   struct Off *p, **p1, *p2;

   /* Symmetric matrix, so r should be > c. */
   if (c > r)
   {
      temp = c;
      c = r;
      r = temp;
   }
   p1 = &BMatrix[r].First;
   p = BMatrix[r].First;
   if (p != NULL)
   {
      /* If we have off-diagonal elements for this row, search through them. */
      while (p->col < c && p->Next)
      {
         p1 = &p->Next;
         p = p->Next;
      }

      /* Element found - add x to value. */
      if (p->col == c)
         p->val += x;
      else
      {
         /* Element not found - allocate new space for it. */
         p2 = AllocElement ();
         p2->val = x;
         p2->col = c;

         /* Link new element into list. */
         if (p->col > c)
         {
            p2->Next = p;
            (*p1) = p2;
         }
         else
         {
            p2->Next = p->Next;
            p->Next = p2;
         }
      }
   }
   else
   {
      /* No off-diagonals for this row yet, so create one. */
      p2 = AllocElement ();
      (*p1) = p2;
      p2->val = x;
      p2->Next = NULL;
      p2->col = c;
   }

   return;
}

/* Calculate likelihood for polygenic trait by integrating out breeding       */
/* values.  Can also fit pseudo-dominance effect (common environment.         */
/* Designed to act as a clone of peel_quant(), except with the ability        */
/* to handle general matrices.                                                */

double gauss_quant (double sigd, double *y, double lambda, int im)
{
   double k, z, nn1, v, lam, c, dt;
   int   i, i1, j, l, t, nneg, size, dom = 0;
   struct Off *p, *p1, *op;
   struct rvec rvec[3];

   if (SigAdd + sigd <= 0.0)
      return envonly (y);
   lam = 1.0;
   if ((im == 1) || (im > 2))
      lam = lambda;

   if (sigd > 0.0)
      dom = 1;

   /* If first time through, set up data structures. */
   if (!recode)
      gauss_setup (dom);
   else if (dom && !fam_list)
      give_up_where (WHERE, ERP_swc);

   /* Number of equations. */
   size = 1 + psize + nfam;

   /* Allocate storage for a matrix for storing the diagonal elements.        */
   BMatrix = malloc (size * sizeof (struct Diag));
   alloc_chk (WHERE, BMatrix);

   /* Also allocate storage for the first block of off-diagonal elements.     */
   if (!BaseBlock)
   {
      BaseBlock = malloc (psize * sizeof (struct Off));
      alloc_chk (WHERE, BaseBlock);
      BaseBlock->Next = NULL;
   }
   bp = 1;
   BBlock = BaseBlock;

   for (i = 0; i < size; i++)
   {
      BMatrix[i].val = 0.0;
      BMatrix[i].First = NULL;
   }

   /* Add in contributions from data. */
   nn1 = k = 0.0;
   c = SigErr + 0.75 * sigd;
   rvec[0].i = 0;
   for (i = 0; i < psize; i++)
      if ((j = data_index[i]))
      {
         nn1++;
         rvec[0].val = -y[j];
         rvec[1].i = eq_pos[i];
         rvec[1].val = lam;
         if (dom)
         {
            j = fam_list[i] + psize - 1;
            rvec[2].i = eq_pos[j];
            rvec[2].val = 1.0;
            i1 = 3;
         }
         else
            i1 = 2;
         for (j = 0; j < i1; j++)
         {
            l = rvec[j].i;
            z = rvec[j].val;
            for (t = 0; t < j; t++)
               AddElement (l, rvec[t].i, z * rvec[t].val / c);
            BMatrix[l].val += z * z / c;
         }
      }
   k -= .5 * nn1 * log (PPI * c);

   /* Add in contributions from additive effect. */
   if (SigAdd > 0.0)
   {
      for (i = 0; i < psize; i++)
      {
         i1 = eq_pos[i];
         BMatrix[i1].val += AIMatrix[i].val / SigAdd;
         for (j = AIMatrix[i].x; j < AIMatrix[i + 1].x; j++)
         {
            AddElement (i1, eq_pos[AIMatrix[j].x], AIMatrix[j].val / SigAdd);
         }
      }
      k -= .5 * (log (PPI * SigAdd) * (double) psize + AIMatrix[psize].val);
   }
   /* Add in contributions from "dominance" effect. */
   if (dom)
   {
      /* Note we are actually using sigd/4. */
      nn1 = 0.0;
      for (i = 0; i < nfam; i++)
      {
         if (BMatrix[eq_pos[i + psize]].val > 0.0)
         {
            BMatrix[eq_pos[i + psize]].val += 4.0 / sigd;
            nn1 += 1.0;
         }
      }
      k -= .5 * nn1 * (log (0.5 * M_PI * sigd));
   }
   /* Now start absorption. */
   nneg = 0;
   dt = nn1 = 0.0;
   for (i = size - 1; i > 0; i--)
   {
      v = BMatrix[i].val;
      if (fabs (v) < 1.0e-16)
         continue;
      if ((p = BMatrix[i].First))
      {
         do
         {
            p1 = BMatrix[i].First;
            while (p != p1)
            {
               AddElement (p->col, p1->col, -p->val * p1->val / v);
               p1 = p1->Next;
            }
            BMatrix[p->col].val -= p->val * p->val / v;
            op = p;
            p = p->Next;
         }
         while (p);
         op->Next = freelist;
         freelist = BMatrix[i].First;
      }
      if (v < 0.0)
      {
         v = -v;
         give_warn (ERP_npv, "gauss_quant", i);
         nneg++;
      }
      dt += log (v);
      nn1 += 1.0;
   }
   if ((p = BMatrix[0].First))
   {
      while (p->Next)
         p = p->Next;
      p->Next = freelist;
      freelist = BMatrix[0].First;
   }
   k += .5 * (nn1 * log (PPI) - dt - BMatrix[0].val);

   if (nneg)
      give_warn (ERP_nnp, "gauss_quant", nneg);
   free (BMatrix);
   BMatrix = NULL;

   return k;
}

void free_BBlock (struct Off *BB)
{
   if (BB->Next)
      free_BBlock (BB->Next);

   free ((void *) BB);
   BB = NULL;

   return;
}

void free_gauss (void)
{
   free ((void *) recode);
   free ((void *) AIMatrix);
   if (fam_list)
      free ((void *) fam_list);
   free ((void *) eq_pos);

   free_BBlock (BBlock);

   return;
}
