/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/*****************************************************************************
  program name: lm_fzplod.c
      (based on new lm_pval in Autozyg; formerly lm_fuzzp)
                                                             EAT Aug 2005

  Program to do posterior IBD probs
  The state (X) consists of segregation indicators

  X determines the IBD pattern among the specified descendant gametes.

  The marker data, Y, is the (unphased) multilocus genotypes of observed 
  individuals in the pedigree.

  The realized probability distributions are scored using a lod score
   for the trait data given a trait model, location, and realized marker
   inheritances.  A multiple-testing corrected version is based on the 
   maximum over loci.

  Output are the fuzzy-pval dsns (per Geyer & Thompson) EAT Oct 2004
          First get null-null dsn quantiles accurate to .001 (0.1%)
          Then score marker loci, as generated, against this dsn.

******************************************************************************/

/**
 * @defgroup lm_markers lm_markers
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_markers"
#define MARKERS
#define CMPNT  0                        /* set =1 for peeling by component */

#include "lodscore.h"

int main (int argc, char *argv[])
{
   int   j, jj;
   int   nitnull, ntp, ntpos;
   int   t, c, h, nsw;
   int ***recounts;                     /* [nc][nl][6] counts of
                                         *             recombinations     */
   double **prob;
   /* likelihoods etc.        */
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;
   int   tot_count = 0;

   /* Map position (cM) of test positions */
   double **rec_left;                   /* fem =0, male=1; by test position   */
   double **rec_rght;                   /* fem =0, male=1; by test position   */
   int  *loc_left;                      /* locus left of test position        */
   int  *loc_rght;                      /* locus rght of test position        */

   double **hdsn;                       /* [ntp+1][nitnull]
                                         * list of null imputation scores
                                         * at all evaluatio positions     */
   double **fzpval;                     /* [npos+1][nnull]             */
   /* summed over cmpt [c]           */

   double **qndsn;                      /* quantiles of the null dsn 
                                           sampled  [npos+1][nnull]     */
   /*  F^{-1}((k+1)/nnull)   is element k     */
   int   nnull;                         /* number of quantiles for null  */
   int   id0;                           /* indicator for id null marginals */
   int   decr;                          /* number of realizations per nnull */

   FILE *oSeedF;

   /* .................................................. */

   nsw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();

   nnull = 1000;

   /* now set evaluation points for lange_lik */
   /* ntp is number of trait positions plus number of markers (NLocPos + 1) */

   alloc_lf_rt (NLocPos + 1, &rec_left, &rec_rght, &loc_left, &loc_rght);

   ntp = set_eval_points (NMarks, loc_left, loc_rght,
                          rec_left[0], rec_rght[0], rec_left[1], rec_rght[1]);
   ntpos = ntp - 1;                     /* number of in-map quntile positions   */
/*  number of null realizations   -- currently set to the same as MCMC Its */
   nitnull = NIt;
/* indicator of id null marginals -- will be 0 for lodscores*/
   id0 = 0;

/*  this will be needed to score the nulls: may cause a problem in setting
    MCMC !! *****    */

   set_locpeel (NMarks);

   /* First generate the null-null distribution quantiles,      EAT  Oct 2004 */
   /*   This will be fixed, and any future sampling of null locus irrelevant  */
   /*    the 0 slot of the hdsn will be used to store the max over loci  */
   /*   For convenience, NIt iterations are used, but iid not mcmc            */

   /* Now realize segs and set hdsn from nitnull nulls */
   hdsn = alloc_dbl2d (ntpos + 1, nitnull);

   printf ("\n now do %d null realizations with ntpos= %d", nitnull, ntpos);
   fflush (stdout);
   for (t = 0; t < nitnull; ++t)
   {
      /* realise the initial switches */

      realise (NMarks, xecomo, nsw);

      /* set founder genome labels (gg0, gg1) from switches */

      for (j = 1; j <= NMarks; ++j)
         set_FGL (j);

/* score the lod score measure -- store the value  */
      do_fzplik (t, ntp, loc_left, loc_rght, rec_left[1], rec_left[0],
                 rec_rght[1], rec_rght[0], hdsn);
   }

/* convert dsn to 0.1 % quantiles */

   qndsn = alloc_dbl2d (ntpos + 1, nnull + 1);

   decr = nitnull / nnull;
   for (j = 0; j <= ntpos; ++j)
      make_quantiles (nnull, decr, hdsn[j], qndsn[j]);

/* free the full set of null dsn realizations: only the quantiles are needed */
   free_2d ((void *) &hdsn);

/* now print quantiles for check -- at some large interval                   */
   printf (" Now print null quantiles\n");
   fflush (stdout);
   pr_quantiles (nnull, nitnull, nnull / 10, ntpos + 1, qndsn);
   fflush (stdout);

/* if marginals should be ident distributed (id0=1), then merge quantiles    */
   if (id0)
   {
      compress_marginals (1, NMarks, nnull, qndsn);

/* now print quantiles again for check -- at some large interval             */
      printf (" Now print merged null quantiles\n");
      fflush (stdout);
      pr_quantiles (nnull, nitnull, nnull / 10, ntpos + 1, qndsn);
      fflush (stdout);
   }

/* allocate and initialize log-seg and log-pen stuff */
/* probabilities now computed separately by component  */

   alloc_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen);

   for (c = 0; c < n_compnts; ++c)
   {
      for (h = 0; h <= NMarks; ++h)
      {
         tot_log_prob_seg[c][h] = 0.0;
         tot_log_prob_pen[c][h] = 0.0;
      }
      for (j = 0; j <= NMarks; ++j)
      {
         for (jj = 0; jj < 6; ++jj)
            recounts[c][j][jj] = 0;
      }
   }

/* allocate and zero the fuzzy-pvals   */
   fzpval = alloc_dbl2d (ntpos + 1, nnull + 1);

/* now intialise inheritance  */
   prob = setupInherit (1);

   alloc_logprob ();

   prntStatusBMC ();
   /* now do burn-in iterations */

   printf (" Now starting %d iterations of burn-in\n", NBrnIt);
   fflush (stdout);
   for (t = 1; t <= NBrnIt; t++)
   {
      if (!(t % ChkIt))
      {
         printf (" Starting burnin iteration %d\n", t);
         fflush (stdout);
      }
      sampling (t, 1, NMarks, nsw, nFGL, prob); /* arg2 =1: EAT 3/2005 */
   }

   /* now do MCMC sampling, and score results */

   printf (" Now starting %d main MCMC iterations \n", NIt);
   fflush (stdout);
   for (t = 0; t < NIt; ++t)
   {
      /* MCMC sampling step or scan */
      sampling (t, 1, NMarks, nsw, nFGL, prob); /* arg2 =1: EAT 3/2005 */

      if (!((t + 1) % ScorIt))
      {
/* now scoring will be as a fuzzy-p dsn for each locus and the max   */
/* If t(S^{(j)}) is exceeded by exactly k of the t(S_0^{(i)}) and tied with w,
 * then contrib to the histogram is rectangle area 1/n, and base endpoints
 *    k/(m+1), (k+1+w)/(m+1)
 */
         scor_fzplik (fzpval, ntp, nnull,
                      loc_left, loc_rght, rec_left[1], rec_left[0], rec_rght[1],
                      rec_rght[0], qndsn);

/* scores log-segs and log-pens       */
         for (c = 0; c < n_compnts; ++c)
         {
            full_seg_prob (NMarks, c, recounts, temp_log_prob_seg);
            for (j = 1; j < NMarks; ++j)
            {
               tot_log_prob_pen[c][j] += log (prob[c][j]);
               tot_log_prob_seg[c][j] += temp_log_prob_seg[c][j];
            }
            tot_log_prob_pen[c][NMarks] += log (prob[c][NMarks]);
         }
         ++tot_count;
      }
   }
   printf ("\n");
   prntStatusFMC ();

   for (j = 0; j <= ntpos; ++j)
      norcum (tot_count, nnull, fzpval[j]);

   printf ("\n    Pointwise distributions of fuzzy p-values, by location:");
   printf ("\npval maxim");

   fflush (stdout);
   pr_pvals (ntpos, fzpval, nnull, nnull / 10);

   for (c = 0; c < n_compnts; ++c)
   {
      pr_cmpnthead (c);
      pr_log_prob_seg (1, NMarks, -cmplist[c + 1].nmeioses, tot_log_prob_seg[c],
                       tot_count);

      pr_log_prob_pen (1, NMarks, tot_log_prob_pen[c], tot_count);

      pr_reco_counts (1, NMarks, recounts[c]);

   }
   printf ("done MCMC\n ");             /* ****  now we are done **** */
   fflush (stdout);

   /* Free up allocated memory */

   /* Free up allocated memory */

   free_xxgg ();

   free_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                &tot_log_prob_pen);
   locus_peel_free ();
   free_allocate_loci ();
   free_chromo (0);
   free_gen_pen ();
   free_seg_pen (NMarks + 1, &prob);
   free_logprob ();
   free_msmpl ();
   free_traits ();
   free_cmplists ();

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
   if (Scan)
      free_perm ();

   exit (EXIT_SUCCESS);
}

/** @} */
